{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "92742f57",
   "metadata": {},
   "source": [
    "# Outlook\n",
    "\n",
    "In this notebook we code the Soft Actor-Critic (SAC) algorithm using BBRL.\n",
    "This algorithm is described in [this\n",
    "paper](http://proceedings.mlr.press/v80/haarnoja18b/haarnoja18b.pdf) and [this\n",
    "paper](https://arxiv.org/pdf/1812.05905.pdf).\n",
    "\n",
    "To understand this code, you need to know more about \n",
    "[the BBRL interaction model](https://github.com/osigaud/bbrl/blob/master/docs/overview.md)\n",
    "Then you should run [a didactical example](https://github.com/osigaud/bbrl/blob/master/docs/notebooks/03-multi_env_autoreset.student.ipynb)\n",
    "to see how agents interact in BBRL when autoreset=True.\n",
    "\n",
    "The algorithm is explained in [this\n",
    "video](https://www.youtube.com/watch?v=U20F-MvThjM) and you can also read [the\n",
    "corresponding slides](http://pages.isir.upmc.fr/~sigaud/teach/ps/12_sac.pdf)."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "4b8f698a",
   "metadata": {},
   "source": [
    "## Installation and Imports\n",
    "\n",
    "### Installation\n",
    "\n",
    "The BBRL library is [here](https://github.com/osigaud/bbrl).\n",
    "\n",
    "We use OmegaConf to that makes it possible that by just defining the `def\n",
    "run_dqn(cfg):` function and then executing a long `params = {...}` variable at\n",
    "the bottom of this colab, the code is run with the parameters without calling\n",
    "an explicit main.\n",
    "\n",
    "More precisely, the code is run by calling\n",
    "\n",
    "`config=OmegaConf.create(params)`\n",
    "\n",
    "`run_dqn(config)`\n",
    "\n",
    "at the very bottom of the colab, after starting tensorboard.\n",
    "\n",
    "Below, we import standard python packages, pytorch packages and gymnasium\n",
    "environments."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "4cc3a2b3",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Installs the necessary Python and system libraries\n",
    "try:\n",
    "    from easypip import easyimport, easyinstall, is_notebook\n",
    "except ModuleNotFoundError as e:\n",
    "    get_ipython().run_line_magic(\"pip\", \"install easypip\")\n",
    "    from easypip import easyimport, easyinstall, is_notebook\n",
    "\n",
    "easyinstall(\"bbrl>=0.2.2\")\n",
    "easyinstall(\"swig\")\n",
    "easyinstall(\"bbrl_gymnasium>=0.2.0\")\n",
    "easyinstall(\"bbrl_gymnasium[box2d]\")\n",
    "easyinstall(\"bbrl_gymnasium[classic_control]\")\n",
    "easyinstall(\"tensorboard\")\n",
    "easyinstall(\"moviepy\")\n",
    "easyinstall(\"box2d-kengz\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9e8408c1",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "import os\n",
    "import sys\n",
    "from pathlib import Path\n",
    "import math\n",
    "\n",
    "from moviepy.editor import ipython_display as video_display\n",
    "import time\n",
    "from tqdm.auto import tqdm\n",
    "from typing import Tuple, Optional\n",
    "from functools import partial\n",
    "\n",
    "from omegaconf import OmegaConf\n",
    "import torch\n",
    "import bbrl_gymnasium\n",
    "\n",
    "import copy\n",
    "from abc import abstractmethod, ABC\n",
    "import torch.nn as nn\n",
    "import torch.nn.functional as F\n",
    "from time import strftime\n",
    "\n",
    "OmegaConf.register_new_resolver(\n",
    "    \"current_time\", lambda: strftime(\"%Y%m%d-%H%M%S\"), replace=True\n",
    ")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a40dae3d",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "lines_to_next_cell": 2,
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Imports all the necessary classes and functions from BBRL\n",
    "from bbrl.agents.agent import Agent\n",
    "from bbrl import get_arguments, get_class, instantiate_class\n",
    "\n",
    "# The workspace is the main class in BBRL, this is where all data is collected and stored\n",
    "from bbrl.workspace import Workspace\n",
    "\n",
    "# Agents(agent1,agent2,agent3,...) executes the different agents the one after the other\n",
    "# TemporalAgent(agent) executes an agent over multiple timesteps in the workspace,\n",
    "# or until a given condition is reached\n",
    "from bbrl.agents import Agents, TemporalAgent\n",
    "\n",
    "# ParallelGymAgent is an agent able to execute a batch of gymnasium environments\n",
    "# with auto-resetting. These agents produce multiple variables in the workspace:\n",
    "# ’env/env_obs’, ’env/reward’, ’env/timestep’, ’env/terminated’,\n",
    "# 'env/truncated', 'env/done', ’env/cumulated_reward’, ...\n",
    "#\n",
    "# When called at timestep t=0, the environments are automatically reset. At\n",
    "# timestep t>0, these agents will read the ’action’ variable in the workspace at\n",
    "# time t − 1\n",
    "from bbrl.agents.gymnasium import GymAgent, ParallelGymAgent, make_env, record_video\n",
    "\n",
    "# Replay buffers are useful to store past transitions when training\n",
    "from bbrl.utils.replay_buffer import ReplayBuffer"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e2e08789",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "# Utility function for launching tensorboard\n",
    "# For Colab - otherwise, it is easier and better to launch tensorboard from\n",
    "# the terminal\n",
    "def setup_tensorboard(path):\n",
    "    path = Path(path)\n",
    "    answer = \"\"\n",
    "    if is_notebook():\n",
    "        if get_ipython().__class__.__module__ == \"google.colab._shell\":\n",
    "            answer = \"y\"\n",
    "        while answer not in [\"y\", \"n\"]:\n",
    "            answer = input(\n",
    "                f\"Do you want to launch tensorboard in this notebook [y/n] \"\n",
    "            ).lower()\n",
    "\n",
    "    if answer == \"y\":\n",
    "        get_ipython().run_line_magic(\"load_ext\", \"tensorboard\")\n",
    "        get_ipython().run_line_magic(\"tensorboard\", f\"--logdir {path.absolute()}\")\n",
    "    else:\n",
    "        import sys\n",
    "        import os\n",
    "        import os.path as osp\n",
    "\n",
    "        print(\n",
    "            f\"Launch tensorboard from the shell:\\n{osp.dirname(sys.executable)}/tensorboard --logdir={path.absolute()}\"\n",
    "        )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e0e96add",
   "metadata": {},
   "source": [
    "### Functions to build networks\n",
    "\n",
    "We define a few utilitary functions to build neural networks"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "32123e76",
   "metadata": {},
   "source": [
    "The function below builds a multi-layer perceptron where the size of each layer is given in the `size` list.\n",
    "We also specify the activation function of neurons at each layer and optionally a different activation function for the final layer."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0db8d462",
   "metadata": {},
   "outputs": [],
   "source": [
    "import torch.nn as nn\n",
    "def build_mlp(sizes, activation, output_activation=nn.Identity()):\n",
    "    \"\"\"Helper function to build a multi-layer perceptron (function from $\\mathbb R^n$ to $\\mathbb R^p$)\n",
    "    \n",
    "    Args:\n",
    "        sizes (List[int]): the number of neurons at each layer\n",
    "        activation (nn.Module): a PyTorch activation function (after each layer but the last)\n",
    "        output_activation (nn.Module): a PyTorch activation function (last layer)\n",
    "    \"\"\"\n",
    "    layers = []\n",
    "    for j in range(len(sizes) - 1):\n",
    "        act = activation if j < len(sizes) - 2 else output_activation\n",
    "        layers += [nn.Linear(sizes[j], sizes[j + 1]), act]\n",
    "    return nn.Sequential(*layers)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b0f99e9d",
   "metadata": {},
   "outputs": [],
   "source": [
    "def build_backbone(sizes, activation):\n",
    "    layers = []\n",
    "    for j in range(len(sizes) - 1):\n",
    "        layers += [nn.Linear(sizes[j], sizes[j + 1]), activation]\n",
    "    return layers"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3f9e0a27",
   "metadata": {},
   "outputs": [],
   "source": [
    "from bbrl import get_arguments, get_class\n",
    "from itertools import chain\n",
    "\n",
    "def setup_optimizer(cfg_optimizer, *agents):\n",
    "    \"\"\"Setup an optimizer for a list of agents\"\"\"\n",
    "    optimizer_args = get_arguments(cfg_optimizer)\n",
    "    parameters = [agent.parameters() for agent in agents]\n",
    "    optimizer = get_class(cfg_optimizer)(chain(*parameters), **optimizer_args)\n",
    "    return optimizer\n",
    "\n",
    "def copy_parameters(model_a, model_b):\n",
    "    \"\"\"Copy parameters from a model a to model_b\"\"\"\n",
    "    for model_a_p, model_b_p in zip(model_a.parameters(), model_b.parameters()):\n",
    "        model_b_p.data.copy_(model_a_p)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "404c6111",
   "metadata": {},
   "source": [
    "### The Logger class\n",
    "\n",
    "The logger is in charge of collecting statistics during the training\n",
    "process.\n",
    "\n",
    "Having logging provided under the hood is one of the features allowing you\n",
    "to save time when using RL libraries like BBRL.\n",
    "\n",
    "In these notebooks, the logger is defined as `bbrl.utils.logger.TFLogger` so as\n",
    "to use a tensorboard visualisation (see the parameters part `params = { \"logger\":{ ...` below).\n",
    "\n",
    "Note that the BBRL Logger is also saving the log in a readable format such\n",
    "that you can use `Logger.read_directories(...)` to read multiple logs, create\n",
    "a dataframe, and analyze many experiments afterward in a notebook for\n",
    "instance. The code for the different kinds of loggers is available in the\n",
    "[bbrl/utils/logger.py](https://github.com/osigaud/bbrl/blob/master/src/bbrl/utils/logger.py)\n",
    "file.\n",
    "\n",
    "`instantiate_class` is an inner BBRL mechanism. The\n",
    "`instantiate_class`function is available in the\n",
    "[`bbrl/__init__.py`](https://github.com/osigaud/bbrl/blob/master/src/bbrl/__init__.py)\n",
    "file."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "de97150e",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "lines_to_next_cell": 2,
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "from bbrl import instantiate_class\n",
    "\n",
    "class Logger():\n",
    "\n",
    "    def __init__(self, cfg):\n",
    "        self.logger = instantiate_class(cfg.logger)\n",
    "\n",
    "    def add_log(self, log_string, loss, steps):\n",
    "        self.logger.add_scalar(log_string, loss.item(), steps)\n",
    "\n",
    "    # A specific function for RL algorithms having a critic, an actor and an entropy losses\n",
    "    def log_losses(self, critic_loss, entropy_loss, actor_loss, steps):\n",
    "        self.add_log(\"critic_loss\", critic_loss, steps)\n",
    "        self.add_log(\"entropy_loss\", entropy_loss, steps)\n",
    "        self.add_log(\"actor_loss\", actor_loss, steps)\n",
    "\n",
    "    def log_reward_losses(self, rewards, nb_steps):\n",
    "        self.add_log(\"reward/mean\", rewards.mean(), nb_steps)\n",
    "        self.add_log(\"reward/max\", rewards.max(), nb_steps)\n",
    "        self.add_log(\"reward/min\", rewards.min(), nb_steps)\n",
    "        self.add_log(\"reward/median\", rewards.median(), nb_steps)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b39be772",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "# Learning environment\n",
    "\n",
    "To setup a common learning environment for RL algorithms, we use the `RLBase`\n",
    "class. This class:\n",
    "\n",
    "1. Initializes the environment (random seed, logger, evaluation environment)\n",
    "2. Defines a `evaluate` method that keeps the best agent so far\n",
    "3. Defines a `visualize_best` method that displays the behavior of the best agent\n",
    "\n",
    "Subclasses need to define `self.train_policy` and `self.eval_policy`, two\n",
    "BBRL agents that respectively choose actions when training and evaluating.\n",
    "\n",
    "The behavior of `RLBase` is controlled by the following configuration\n",
    "variables:\n",
    "\n",
    "- `base_dir` defines the directory subpath used when outputing losses during\n",
    "training as well as other outputs (serialized agent, global statistics, etc.)\n",
    "- `algorithm.seed` defines the random seed used (to initialize the agent and\n",
    "  the environment)\n",
    "- `gym_env` defines the gymnasium environment, and in particular\n",
    "`gym_env.env_name` the name of the gymansium environment\n",
    "- `logger` defines what type of logger is used to log the different values\n",
    "associated with learning\n",
    "- `algorithm.eval_interval` defines the number of observed transitions between\n",
    "each evaluation of the agent"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e97199bd",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "from typing import Any\n",
    "import logging\n",
    "from abc import ABC\n",
    "from functools import cached_property\n",
    "\n",
    "class RLBase(ABC):\n",
    "    \"\"\"Base class for Reinforcement learning algorithms\n",
    "    \n",
    "    This class deals with common processing:\n",
    "\n",
    "    - defines the logger, the train and evaluation agents\n",
    "    - defines how to evaluate a policy\n",
    "    \"\"\"\n",
    "\n",
    "    #: The configuration\n",
    "    cfg: Any\n",
    "\n",
    "    #: The evaluation environment deals with the last action, and produces a new\n",
    "    # state of the environment\n",
    "    eval_env: Agent\n",
    "\n",
    "    #: The training policy\n",
    "    train_policy: Agent\n",
    "\n",
    "    #: The evaluation policy (if not defined, uses the training policy)\n",
    "    eval_policy: Agent\n",
    "\n",
    "    def __init__(self, cfg):\n",
    "        # Basic initialization\n",
    "        self.cfg = cfg\n",
    "        torch.manual_seed(cfg.algorithm.seed)\n",
    "\n",
    "        # Sets the base directory and logger directory\n",
    "        base_dir = Path(self.cfg.base_dir)\n",
    "        self.base_dir = Path(\"outputs\") / Path(self.cfg.base_dir)\n",
    "        self.base_dir.mkdir(parents=True, exist_ok=True)\n",
    "\n",
    "        # Initialize the logger class\n",
    "        if not hasattr(cfg.logger, \"log_dir\"):\n",
    "            cfg.logger.log_dir = str(Path(\"outputs\") / \"tblogs\" / base_dir)\n",
    "        self.logger = Logger(cfg)\n",
    "\n",
    "        # Subclasses have to define the training and eval policies\n",
    "        self.train_policy = None\n",
    "        self.eval_policy = None\n",
    "\n",
    "        # Sets up the evaluation environment\n",
    "        self.eval_env = ParallelGymAgent(\n",
    "            partial(make_env, cfg.gym_env.env_name), \n",
    "            cfg.algorithm.nb_evals\n",
    "        ).seed(cfg.algorithm.seed)\n",
    "\n",
    "        # Initialize values\n",
    "        self.last_eval_step = 0\n",
    "        self.nb_steps = 0\n",
    "        self.best_policy = None\n",
    "        self.best_reward = -torch.inf\n",
    "\n",
    "        # Records the rewards\n",
    "        self.eval_rewards = []\n",
    "\n",
    "    @cached_property\n",
    "    def train_agent(self):\n",
    "        \"\"\"Returns the training agent\n",
    "\n",
    "        The agent is composed of a policy agent and the training environment.      \n",
    "        This method supposes that `self.train_policy` has been setup\n",
    "        \"\"\"\n",
    "        assert self.train_policy is not None, \"The train_policy property is not defined before the policy is set\"\n",
    "        return TemporalAgent(Agents(self.train_env, self.train_policy))\n",
    "\n",
    "    @cached_property\n",
    "    def eval_agent(self):\n",
    "        \"\"\"Returns the evaluation agent \n",
    "        \n",
    "        The agent is composed of a policy agent and the evaluation environment\n",
    "        \n",
    "        Uses `self.eval_policy` (or `self.train_policy` if not defined)\n",
    "\n",
    "        \"\"\"\n",
    "        assert self.eval_policy is not None or self.train_policy is not None, \"eval_agent property is not defined before the policy is set\"\n",
    "        return TemporalAgent(Agents(self.eval_env, self.eval_policy if self.eval_policy is not None else self.train_policy))\n",
    "\n",
    "    def evaluate(self):\n",
    "        \"\"\"Evaluate the current policy `self.eval_policy`\n",
    "        \n",
    "        Evaluation is conducted every `cfg.algorithm.eval_interval` steps, and\n",
    "        we keep a copy of the best agent so far in `self.best_policy`\n",
    "        \n",
    "        Returns True if the current policy is the best so far\n",
    "        \"\"\"\n",
    "        if (self.nb_steps - self.last_eval_step) > self.cfg.algorithm.eval_interval:\n",
    "            self.last_eval_step = self.nb_steps\n",
    "            eval_workspace = Workspace() \n",
    "            self.eval_agent(\n",
    "                eval_workspace,\n",
    "                t=0,\n",
    "                stop_variable=\"env/done\"\n",
    "            )\n",
    "            rewards = eval_workspace[\"env/cumulated_reward\"][-1]\n",
    "            self.logger.log_reward_losses(rewards, self.nb_steps)\n",
    "\n",
    "            if getattr(self.cfg, \"collect_stats\", False):\n",
    "                self.eval_rewards.append(rewards)\n",
    "\n",
    "            rewards_mean = rewards.mean()\n",
    "            if rewards_mean > self.best_reward:\n",
    "                self.best_policy = copy.deepcopy(self.eval_policy)\n",
    "                self.best_reward = rewards_mean\n",
    "                return True\n",
    "\n",
    "    def save_stats(self):\n",
    "        \"\"\"Save reward statistics into `stats.npy`\"\"\"\n",
    "        if getattr(self.cfg, \"collect_stats\", False) and self.eval_rewards:\n",
    "            data = torch.stack(self.eval_rewards, axis=-1) \n",
    "            with (self.base_dir / \"stats.npy\").open(\"wt\") as fp:\n",
    "                np.savetxt(fp, data.numpy())\n",
    "\n",
    "    def visualize_best(self):\n",
    "        \"\"\"Visualize the best agent\"\"\"\n",
    "        env = make_env(self.cfg.gym_env.env_name, render_mode=\"rgb_array\")\n",
    "        path = self.base_dir / \"best_agent.mp4\"\n",
    "        print(f\"Video of best agent recorded in {path}\")\n",
    "        record_video(env, self.best_policy, path)\n",
    "        return video_display(str(path.absolute()))"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "dfd17b77",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "The `EpochBasedAlgo` defines the environment when using replay buffers. In\n",
    "particular, it defines `self.train_env` which is the environment used for\n",
    "training. This environment uses *autoreset*, i.e. when reaching a terminal\n",
    "state, a new environment is created.\n",
    "\n",
    "`EpochBasedAlgo` stores all the transitions $(s_t, a_t, r_t, s_{t+1}, ...)$\n",
    "into `self.replay_buffer`.\n",
    "\n",
    "The behavior of `EpochBasedAlgo` is controlled by the following configuration\n",
    "variables:\n",
    "\n",
    "- `gym_env.env_name` defines the gymnasium environment\n",
    "- `algorithm.n_envs` defines the number of parallel environments\n",
    "- `algorithm.seed` defines the random seed used (to initialize the agent and\n",
    "  the environment)\n",
    "- `algorithm.buffer_size` is the maximum number of transitions stored into the\n",
    "replay buffer"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "55e17441",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "class EpochBasedAlgo(RLBase):\n",
    "    \"\"\"RL environment when using transition buffers\"\"\"\n",
    "    \n",
    "    train_agent: TemporalAgent\n",
    "    \n",
    "    \"\"\"Base class for RL experiments with full episodes\"\"\"\n",
    "    def __init__(self, cfg):\n",
    "        super().__init__(cfg)\n",
    "\n",
    "        # We use a non-autoreset workspace\n",
    "        self.train_env = ParallelGymAgent(\n",
    "            partial(make_env, cfg.gym_env.env_name, autoreset=True), \n",
    "            cfg.algorithm.n_envs\n",
    "        ).seed(cfg.algorithm.seed)\n",
    "\n",
    "        # Configure the workspace to the right dimension\n",
    "        # Note that no parameter is needed to create the workspace.\n",
    "        self.replay_buffer = ReplayBuffer(max_size=cfg.algorithm.buffer_size)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "ad73d393",
   "metadata": {},
   "source": [
    "`iter_replay_buffers` provides an easy access to the replay buffer when\n",
    "learning. Its behavior depends on several configuration values:\n",
    "\n",
    "- `cfg.algorithm.max_epochs` defines the number of times the agent is used to\n",
    "collect transitions\n",
    "- `cfg.algorithm.learning_starts` defines the number of transitions before\n",
    "learning starts\n",
    " \n",
    "Using `iter_replay_buffers` is simple:\n",
    "\n",
    "```py\n",
    "  class MyAlgo(EpochBasedAlgo):\n",
    "      def __init__(self, cfg):\n",
    "          super().__init__(cfg)\n",
    "\n",
    "          # Define the train and evaluation policies\n",
    "          # (the agents compute the workspace `action` variable)\n",
    "          self.train_policy = ...\n",
    "          self.eval_policy = ...\n",
    "\n",
    "  rl_algo = MyAlgo(cfg)\n",
    "  for rb in iter_replay_buffers(rl_algo):\n",
    "      # rb is a workspace containing transitions\n",
    "      ...\n",
    "```"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c5910d92",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "def iter_replay_buffers(algo: EpochBasedAlgo):\n",
    "    \"\"\"Loop over transition buffers\"\"\"\n",
    "    train_workspace = Workspace()\n",
    "\n",
    "    epochs_pb = tqdm(range(algo.cfg.algorithm.max_epochs))\n",
    "    for epoch in epochs_pb:\n",
    "        \n",
    "        # This is the tricky part with transition buffers. The difficulty lies in the\n",
    "        # copy of the last step and the way to deal with the n_steps return.\n",
    "        #\n",
    "        # The call to `train_agent(workspace, t=1, n_steps=cfg.algorithm.n_timesteps -\n",
    "        # 1, stochastic=True)` makes the agent run a number of steps in the workspace.\n",
    "        # In practice, it calls the\n",
    "        # [`__call__(...)`](https://github.com/osigaud/bbrl/blob/master/src/bbrl/agents/agent.py#L59)\n",
    "        # function which makes a forward pass of the agent network using the workspace\n",
    "        # data and updates the workspace accordingly.\n",
    "        #\n",
    "        # Now, if we start at the first epoch (`epoch=0`), we start from the first step\n",
    "        # (`t=0`). But when subsequently we perform the next epochs (`epoch>0`), we must\n",
    "        # not forget to cover the transition at the border between the previous epoch\n",
    "        # and the current epoch. To avoid this risk, we copy the information from the\n",
    "        # last time step of the previous epoch into the first time step of the next\n",
    "        # epoch. This is explained in more details in [a previous\n",
    "        # notebook](https://colab.research.google.com/drive/1W9Y-3fa6LsPeR6cBC1vgwBjKfgMwZvP5).\n",
    "        if epoch == 0:\n",
    "            # First run: we start from scratch\n",
    "            algo.train_agent(\n",
    "                train_workspace, t=0, n_steps=algo.cfg.algorithm.n_steps, stochastic=True\n",
    "            )\n",
    "        else:\n",
    "            # Other runs: we copy the last step and start from there\n",
    "            train_workspace.zero_grad()\n",
    "            train_workspace.copy_n_last_steps(1)\n",
    "            algo.train_agent(\n",
    "                train_workspace, t=1, n_steps=algo.cfg.algorithm.n_steps-1, stochastic=True\n",
    "            )\n",
    "        \n",
    "        algo.nb_steps += algo.cfg.algorithm.n_steps * algo.cfg.algorithm.n_envs\n",
    "\n",
    "        # Add transitions to buffer\n",
    "        transition_workspace = train_workspace.get_transitions()\n",
    "        algo.replay_buffer.put(transition_workspace)\n",
    "        if algo.replay_buffer.size() > algo.cfg.algorithm.learning_starts:\n",
    "            yield algo.replay_buffer\n",
    "\n",
    "        # Eval\n",
    "        epochs_pb.set_description(\n",
    "            f\"nb_steps: {algo.nb_steps}, \"\n",
    "            f\"best reward: {algo.best_reward:.2f}\"\n",
    "        )"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7fc3b712",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "### Soft parameter updates\n",
    "\n",
    "To update the target critic, one uses the following equation:\n",
    "$\\theta' \\leftarrow \\tau \\theta + (1- \\tau) \\theta'$\n",
    "where $\\theta$ is the vector of parameters of the critic, and $\\theta'$ is the vector of parameters of the target critic.\n",
    "The `soft_update_params(...)` function is in charge of performing this soft update."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "4c7aa85d",
   "metadata": {
    "jupyter": {
     "source_hidden": true
    },
    "tags": [
     "hide-input"
    ]
   },
   "outputs": [],
   "source": [
    "def soft_update_params(net, target_net, tau):\n",
    "    for param, target_param in zip(net.parameters(), target_net.parameters()):\n",
    "        target_param.data.copy_(tau * param.data + (1 - tau) * target_param.data)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "42c4981e",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "## The SquashedGaussianActor\n",
    "\n",
    "SAC works better with a Squashed Gaussian actor, which enables the\n",
    "reparametrization trick. Note that our attempts to use a\n",
    "`TunableVarianceContinuousActor` as we did for instance in the [notebook about\n",
    "PPO](http://master-dac.isir.upmc.fr/rld/rl/07-1-ppo_penalty.student.ipynb)\n",
    "completely failed. Such failure is also documented in the [OpenAI spinning up\n",
    "documentation page about\n",
    "SAC](https://spinningup.openai.com/en/latest/algorithms/sac.html).\n",
    "\n",
    "The code of the `SquashedGaussianActor` actor is below.\n",
    "\n",
    "The fact that we use the reparametrization trick is hidden inside the code of\n",
    "this distribution. "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "16a197aa",
   "metadata": {},
   "outputs": [],
   "source": [
    "from torch.distributions import Normal, Independent, TransformedDistribution, TanhTransform\n",
    "\n",
    "class SquashedGaussianActor(Agent):\n",
    "    def __init__(self, state_dim, hidden_layers, action_dim, min_std=1e-4):\n",
    "        \"\"\"Creates a new Squashed Gaussian actor\n",
    "\n",
    "        :param state_dim: The dimension of the state space\n",
    "        :param hidden_layers: Hidden layer sizes\n",
    "        :param action_dim: The dimension of the action space\n",
    "        :param min_std: The minimum standard deviation, defaults to 1e-4\n",
    "        \"\"\"\n",
    "        super().__init__()\n",
    "        self.min_std = min_std\n",
    "        backbone_dim = [state_dim] + list(hidden_layers)\n",
    "        self.layers = build_backbone(backbone_dim, activation=nn.ReLU())\n",
    "        self.backbone = nn.Sequential(*self.layers)\n",
    "        self.last_mean_layer = nn.Linear(hidden_layers[-1], action_dim)\n",
    "        self.last_std_layer = nn.Linear(hidden_layers[-1], action_dim)\n",
    "        # std must be positive\n",
    "        self.softplus = nn.Softplus()\n",
    "        # cache_size avoids numerical infinites or NaNs when\n",
    "        # computing log probabilities\n",
    "        self.tanh_transform = TanhTransform(cache_size=1)\n",
    "\n",
    "    def normal_dist(self, obs: torch.Tensor):\n",
    "        # Compute normal distribution given observation(s)\n",
    "        backbone_output = self.backbone(obs)\n",
    "        mean = self.last_mean_layer(backbone_output)\n",
    "        std_out = self.last_std_layer(backbone_output)\n",
    "        std = self.softplus(std_out) + self.min_std\n",
    "        # Independent ensures that we have a multivariate \n",
    "        # Gaussian with a diagonal covariance matrix (given as\n",
    "        # a vector `std`)\n",
    "        return Independent(Normal(mean, std), 1)\n",
    "\n",
    "    def forward(self, t, stochastic=True):\n",
    "        normal_dist = self.normal_dist(self.get((\"env/env_obs\", t)))\n",
    "        action_dist = TransformedDistribution(normal_dist, [self.tanh_transform])\n",
    "        if stochastic:\n",
    "            # Uses the re-parametrization trick\n",
    "            action = action_dist.rsample()\n",
    "        else:\n",
    "            action = self.tanh_transform(normal_dist.mode)\n",
    "            \n",
    "        log_prob = action_dist.log_prob(action)\n",
    "        # This line allows to deepcopy the actor...\n",
    "        self.tanh_transform._cached_x_y = [None, None]\n",
    "        self.set((\"action\", t), action)\n",
    "        self.set((\"action_logprobs\", t), log_prob)\n",
    "    \n",
    "class DeterministicWrapper(Agent):\n",
    "    \"\"\"Use the above agent with a deterministic policy\"\"\"\n",
    "    def __init__(self, actor: SquashedGaussianActor):\n",
    "        super().__init__()\n",
    "        self.actor = actor\n",
    "        \n",
    "    def forward(self, t, **args):\n",
    "        self.actor(self.workspace, t=t, stochastic=False)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "5e307691",
   "metadata": {},
   "source": [
    "### CriticAgent\n",
    "\n",
    "As critics and target critics, SAC uses several instances of ContinuousQAgent\n",
    "class, as DDPG and TD3.\n",
    "See the [DDPG notebook](http://master-dac.isir.upmc.fr/rld/rl/04-ddpg-td3.student.ipynb) for details."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "33f378d1",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "class ContinuousQAgent(Agent):\n",
    "    def __init__(self, state_dim, hidden_layers, action_dim):\n",
    "        super().__init__()\n",
    "        self.is_q_function = True\n",
    "        self.model = build_mlp(\n",
    "            [state_dim + action_dim] + list(hidden_layers) + [1], activation=nn.ReLU()\n",
    "        )\n",
    "\n",
    "    def forward(self, t):\n",
    "        obs = self.get((\"env/env_obs\", t))\n",
    "        action = self.get((\"action\", t))\n",
    "        obs_act = torch.cat((obs, action), dim=1)\n",
    "        q_value = self.model(obs_act).squeeze(-1)\n",
    "        self.set((f\"{self.prefix}q_value\", t), q_value)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "27166e5d",
   "metadata": {},
   "source": [
    "### Building the complete training and evaluation agents\n",
    "\n",
    "In the code below we create the Squashed Gaussian actor, two critics and the\n",
    "corresponding target critics. Beforehand, we checked that the environment\n",
    "takes continuous actions (otherwise we would need a different code)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9b6dfad2",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Create the SAC Agent\n",
    "class SACAgents(EpochBasedAlgo):\n",
    "    def __init__(self, cfg):\n",
    "        super().__init__(cfg)\n",
    "        \n",
    "        obs_size, act_size = self.train_env.get_obs_and_actions_sizes()\n",
    "        assert (\n",
    "            self.train_env.is_continuous_action()\n",
    "        ), \"SAC code dedicated to continuous actions\"\n",
    "\n",
    "        # We need an actor\n",
    "        self.actor= SquashedGaussianActor(\n",
    "            obs_size, cfg.algorithm.architecture.actor_hidden_size, act_size\n",
    "        )\n",
    "\n",
    "        # Builds the critics\n",
    "        self.critic_1 = ContinuousQAgent(\n",
    "            obs_size, cfg.algorithm.architecture.critic_hidden_size, act_size, \n",
    "        ).with_prefix(\"critic-1/\")\n",
    "        self.target_critic_1 = copy.deepcopy(self.critic_1).with_prefix(\"target-critic-1/\")\n",
    "        \n",
    "        self.critic_2 = ContinuousQAgent(\n",
    "            obs_size, cfg.algorithm.architecture.critic_hidden_size, act_size,\n",
    "        ).with_prefix(\"critic-2/\")\n",
    "        self.target_critic_2 = copy.deepcopy(self.critic_2).with_prefix(\"target-critic-2/\")\n",
    "        \n",
    "        # Train and evaluation policies\n",
    "        self.train_policy = self.actor\n",
    "        self.eval_policy = DeterministicWrapper(self.actor)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "10187409",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "For the entropy coefficient optimizer, the code is as follows. Note the trick\n",
    "which consists in using the log of this entropy coefficient. This trick was\n",
    "taken from the Stable baselines3 implementation of SAC, which is explained in\n",
    "[this\n",
    "notebook](https://colab.research.google.com/drive/12LER1_ShWOa_UhOL1nlX-LX_t5KQK9LV?usp=sharing).\n",
    "\n",
    "Tuning $\\alpha$ in SAC is an option. To chose to tune it, the `target_entropy`\n",
    "argument in the parameters should be `auto`. The initial value is given\n",
    "through the `entropy_coef` parameter. For any other value than `auto`, the\n",
    "value of $\\alpha$ will stay constant and correspond to the `entropy_coef`\n",
    "parameter."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "77dac2ed",
   "metadata": {},
   "outputs": [],
   "source": [
    "def setup_entropy_optimizers(cfg):\n",
    "    if cfg.algorithm.entropy_mode == \"auto\":\n",
    "        entropy_coef_optimizer_args = get_arguments(cfg.entropy_coef_optimizer)\n",
    "        # Note: we optimize the log of the entropy coef which is slightly different from the paper\n",
    "        # as discussed in https://github.com/rail-berkeley/softlearning/issues/37\n",
    "        # Comment and code taken from the SB3 version of SAC\n",
    "        log_entropy_coef = torch.log(\n",
    "            torch.ones(1) * cfg.algorithm.init_entropy_coef\n",
    "        ).requires_grad_(True)\n",
    "        entropy_coef_optimizer = get_class(cfg.entropy_coef_optimizer)(\n",
    "            [log_entropy_coef], **entropy_coef_optimizer_args\n",
    "        )\n",
    "        return entropy_coef_optimizer, log_entropy_coef\n",
    "    else:\n",
    "        return None, None"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "6847a402",
   "metadata": {
    "lines_to_next_cell": 2
   },
   "source": [
    "### Compute the critic loss\n",
    "\n",
    "With the notations of my slides, the equation corresponding to Eq. (5) and (6)\n",
    "in [this paper](https://arxiv.org/pdf/1812.05905.pdf) becomes:\n",
    "\n",
    "$$ loss_Q({\\boldsymbol{\\theta}}) = {\\rm I\\!E}_{(\\mathbf{s}_t, \\mathbf{a}_t,\n",
    "\\mathbf{s}_{t+1}) \\sim \\mathcal{D}}\\left[\\left( r(\\mathbf{s}_t, \\mathbf{a}_t)\n",
    "+ \\gamma {\\rm I\\!E}_{\\mathbf{a} \\sim\n",
    "\\pi_{\\boldsymbol{\\theta}}(.|\\mathbf{s}_{t+1})}\\left[\\hat{Q}^{\\pi_{\\boldsymbol{\\theta}}}_{\\boldsymbol{\\phi}}(\\mathbf{s}_{t+1},\n",
    "\\mathbf{a}) - \\alpha\n",
    "\\log{\\pi_{\\boldsymbol{\\theta}}(\\mathbf{a}|\\mathbf{s}_{t+1})} \\right] -\n",
    "\\hat{Q}^{\\pi_{\\boldsymbol{\\theta}}}_{\\boldsymbol{\\phi}}(\\mathbf{s}_t,\n",
    "\\mathbf{a}_t) \\right)^2 \\right] $$\n",
    "\n",
    "An important information in the above equation and the one about the actor\n",
    "loss below is the index of the expectations. These indexes tell us where the\n",
    "data should be taken from. In the above equation, one can see that the index\n",
    "of the outer expectation is over samples taken from the replay buffer, whereas\n",
    "in the inner expectation we consider actions from the current actor at the\n",
    "next state.\n",
    "\n",
    "Thus, to compute the inner expectation, one needs to determine what actions\n",
    "the current actor would take in the next state of each sample. This is what\n",
    "the line \n",
    "\n",
    "`t_actor(rb_workspace, t=1, n_steps=1, stochastic=True)`\n",
    "\n",
    "does. The parameter `t=1` (instead of 0) ensures that we consider the next\n",
    "state.\n",
    "\n",
    "Once we have determined these actions, we can determine their Q-values and\n",
    "their log probabilities, to compute the inner expectation.\n",
    "\n",
    "Note that at this stage, we only determine the log probabilities corresponding\n",
    "to actions taken at the next time step, by contrast with what we do for the\n",
    "actor in the `compute_actor_loss(...)` function later on.\n",
    "\n",
    "Finally, once we have computed the $$\n",
    "\\hat{Q}^{\\pi_{\\boldsymbol{\\theta}}}_{\\boldsymbol{\\phi}}(\\mathbf{s}_{t+1},\n",
    "\\mathbf{a}) $$ for both critics, we take the min and store it into\n",
    "`post_q_values`. By contrast, the Q-values corresponding to the last term of\n",
    "the equation are taken from the replay buffer, they are computed in the\n",
    "beginning of the function by applying the Q agents to the replay buffer\n",
    "*before* changing the action to that of the current actor.\n",
    "\n",
    "An important remark is that, if the entropy coefficient $\\alpha$ corresponding\n",
    "to the `ent_coef` variable is set to 0, then we retrieve exactly the critic\n",
    "loss computation function of the TD3 algorithm. As we will see later, this is\n",
    "also true of the actor loss computation.\n",
    "\n",
    "This remark proved very useful in debugging the SAC code. We have set\n",
    "`ent_coef` to 0 and ensured the behavior was strictly the same as the behavior\n",
    "of TD3."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "2f68ab76",
   "metadata": {
    "lines_to_next_cell": 1
   },
   "outputs": [],
   "source": [
    "def compute_critic_loss(\n",
    "    cfg, reward, must_bootstrap,\n",
    "    t_actor, \n",
    "    q_agents, \n",
    "    target_q_agents, \n",
    "    rb_workspace,\n",
    "    ent_coef\n",
    "):\n",
    "    \"\"\"Computes the critic loss for a set of $S$ transition samples\n",
    "\n",
    "    Args:\n",
    "        cfg: The experimental configuration\n",
    "        reward: Tensor (2xS) of rewards\n",
    "        must_bootstrap: Tensor (S) of indicators\n",
    "        t_actor: The actor agent (as a TemporalAgent)\n",
    "        q_agents: The critics (as a TemporalAgent)\n",
    "        target_q_agents: The target of the critics (as a TemporalAgent)\n",
    "        rb_workspace: The transition workspace\n",
    "        ent_coef: The entropy coefficient\n",
    "\n",
    "    Returns:\n",
    "        Tuple[torch.Tensor, torch.Tensor]: The two critic losses (scalars)\n",
    "    \"\"\"\n",
    "\n",
    "    # Compute q_values from both critics with the actions present in the buffer:\n",
    "    # at t, we have Q(s,a) from the (s,a) in the RB\n",
    "    q_agents(rb_workspace, t=0, n_steps=1)\n",
    "\n",
    "    with torch.no_grad():\n",
    "        # Replay the current actor on the replay buffer to get actions of the\n",
    "        # current actor\n",
    "        t_actor(rb_workspace, t=1, n_steps=1, stochastic=True)\n",
    "        action_logprobs_next = rb_workspace[\"action_logprobs\"]\n",
    "\n",
    "        # Compute target q_values from both target critics: at t+1, we have\n",
    "        # Q(s+1,a+1) from the (s+1,a+1) where a+1 has been replaced in the RB\n",
    "        target_q_agents(rb_workspace, t=1, n_steps=1)\n",
    "\n",
    "    q_values_rb_1, q_values_rb_2, post_q_values_1, post_q_values_2 = rb_workspace[\n",
    "        \"critic-1/q_value\", \"critic-2/q_value\", \"target-critic-1/q_value\", \"target-critic-2/q_value\"\n",
    "    ]\n",
    "\n",
    "    # Compute temporal difference\n",
    "\n",
    "    assert False, 'Not implemented yet'\n",
    "\n",
    "\n",
    "    return critic_loss_1, critic_loss_2"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a8d72e02",
   "metadata": {},
   "source": [
    "### Compute the actor Loss\n",
    "\n",
    "With the notations of my slides, the equation of the actor loss corresponding\n",
    "to Eq. (7) in [this paper](https://arxiv.org/pdf/1812.05905.pdf) becomes:\n",
    "\n",
    "$$ loss_\\pi({\\boldsymbol{\\theta}}) = {\\rm I\\!E}_{\\mathbf{s}_t \\sim\n",
    "\\mathcal{D}}\\left[ {\\rm I\\!E}_{\\mathbf{a}_t\\sim\n",
    "\\pi_{\\boldsymbol{\\theta}}(.|\\mathbf{s}_t)} \\left[ \\alpha\n",
    "\\log{\\pi_{\\boldsymbol{\\theta}}(\\mathbf{a}_t|\\mathbf{s}_t) -\n",
    "\\hat{Q}^{\\pi_{\\boldsymbol{\\theta}}}_{\\boldsymbol{\\phi}}(\\mathbf{s}_t,\n",
    "\\mathbf{a}_t)} \\right] \\right] $$\n",
    "\n",
    "Note that [the paper](https://arxiv.org/pdf/1812.05905.pdf) mistakenly writes\n",
    "$Q_\\theta(s_t,s_t)$\n",
    "\n",
    "As for the critic loss, we have two expectations, one over the states from the\n",
    "replay buffer, and one over the actions of the current actor. Thus we need to\n",
    "apply again the current actor to the content of the replay buffer.\n",
    "\n",
    "But this time, we consider the current state, thus we parametrize it with\n",
    "`t=0` and `n_steps=1`. This way, we get the log probabilities and Q-values at\n",
    "the current step.\n",
    "\n",
    "A nice thing is that this way, there is no overlap between the log probability\n",
    "data used to update the critic and the actor, which avoids having to 'retain'\n",
    "the computation graph so that it can be reused for the actor and the critic.\n",
    "\n",
    "This small trick is one of the features that makes coding SAC the most\n",
    "difficult.\n",
    "\n",
    "Again, once we have computed the Q values over both critics, we take the min\n",
    "and put it into `current_q_values`.\n",
    "\n",
    "As for the critic loss, if we set `ent_coef` to 0, we retrieve the actor loss\n",
    "function of DDPG and TD3, which simply tries to get actions that maximize the\n",
    "Q values (by minimizing -Q)."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9734e09c",
   "metadata": {},
   "outputs": [],
   "source": [
    "def compute_actor_loss(ent_coef, t_actor, q_agents, rb_workspace):\n",
    "    \"\"\"\n",
    "    Actor loss computation\n",
    "    :param ent_coef: The entropy coefficient $\\alpha$\n",
    "    :param t_actor: The actor agent (temporal agent)\n",
    "    :param q_agents: The critics (as temporal agent)\n",
    "    :param rb_workspace: The replay buffer (2 time steps, $t$ and $t+1$)\n",
    "    \"\"\"\n",
    "    \n",
    "    # Recompute the q_values from the current actor, not from the actions in the buffer\n",
    "\n",
    "    # Recompute the action with the current actor (at $a_t$)\n",
    "\n",
    "    assert False, 'Not implemented yet'\n",
    "\n",
    "\n",
    "    # Compute Q-values\n",
    "\n",
    "    assert False, 'Not implemented yet'\n",
    "\n",
    "    current_q_values = torch.min(q_values_1, q_values_2)\n",
    "\n",
    "    # Compute the actor loss\n",
    "\n",
    "    actor_loss =\n",
    "    assert False, 'Not implemented yet'\n",
    "\n",
    "\n",
    "    return actor_loss.mean()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "6fe7f804",
   "metadata": {},
   "source": [
    "## Main training loop"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "22160566",
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "\n",
    "def run_sac(agents: SACAgents):\n",
    "    cfg = agents.cfg\n",
    "    logger = agents.logger\n",
    "\n",
    "    # init_entropy_coef is the initial value of the entropy coef alpha. \n",
    "    ent_coef = cfg.algorithm.init_entropy_coef\n",
    "    tau = cfg.algorithm.tau_target\n",
    "\n",
    "    # Creates the temporal actors\n",
    "    t_actor = TemporalAgent(agents.train_policy)\n",
    "    q_agents = TemporalAgent(Agents(agents.critic_1, agents.critic_2))\n",
    "    target_q_agents = TemporalAgent(Agents(agents.target_critic_1, agents.target_critic_2))\n",
    "\n",
    "    # Configure the optimizer\n",
    "    actor_optimizer = setup_optimizer(cfg.actor_optimizer, agents.actor)\n",
    "    critic_optimizer = setup_optimizer(cfg.critic_optimizer, agents.critic_1, agents.critic_2)\n",
    "    entropy_coef_optimizer, log_entropy_coef = setup_entropy_optimizers(cfg)\n",
    "\n",
    "    # If entropy_mode is not auto, the entropy coefficient ent_coef will remain fixed\n",
    "    if cfg.algorithm.entropy_mode == \"auto\":\n",
    "        # target_entropy is \\mathcal{H}_0 in the SAC and aplications paper. \n",
    "        target_entropy = -np.prod(agents.train_env.action_space.shape).astype(np.float32)\n",
    "\n",
    "    for rb in iter_replay_buffers(agents):\n",
    "        rb_workspace = rb.get_shuffled(cfg.algorithm.batch_size)\n",
    "\n",
    "        terminated, reward = rb_workspace[\n",
    "            \"env/terminated\", \"env/reward\"\n",
    "        ]\n",
    "        if entropy_coef_optimizer is not None:\n",
    "            ent_coef = torch.exp(log_entropy_coef.detach())\n",
    "\n",
    "        # Critic update part ###############################\n",
    "        critic_optimizer.zero_grad()\n",
    "\n",
    "        (\n",
    "            critic_loss_1, critic_loss_2\n",
    "        ) = compute_critic_loss(\n",
    "            cfg, \n",
    "            reward, \n",
    "            ~terminated[1],\n",
    "            t_actor,\n",
    "            q_agents,\n",
    "            target_q_agents,\n",
    "            rb_workspace,\n",
    "            ent_coef\n",
    "        )\n",
    "\n",
    "        logger.add_log(\"critic_loss_1\", critic_loss_1, agents.nb_steps)\n",
    "        logger.add_log(\"critic_loss_2\", critic_loss_2, agents.nb_steps)\n",
    "        critic_loss = critic_loss_1 + critic_loss_2\n",
    "        critic_loss.backward()\n",
    "        torch.nn.utils.clip_grad_norm_(\n",
    "            agents.critic_1.parameters(), cfg.algorithm.max_grad_norm\n",
    "        )\n",
    "        torch.nn.utils.clip_grad_norm_(\n",
    "            agents.critic_2.parameters(), cfg.algorithm.max_grad_norm\n",
    "        )\n",
    "        critic_optimizer.step()\n",
    "\n",
    "\n",
    "        # Actor update part ###############################\n",
    "        actor_optimizer.zero_grad()\n",
    "        actor_loss = compute_actor_loss(\n",
    "            ent_coef, t_actor, q_agents, rb_workspace\n",
    "        )\n",
    "        logger.add_log(\"actor_loss\", actor_loss, agents.nb_steps)\n",
    "        actor_loss.backward()\n",
    "        torch.nn.utils.clip_grad_norm_(\n",
    "            agents.actor.parameters(), cfg.algorithm.max_grad_norm\n",
    "        )\n",
    "        actor_optimizer.step()\n",
    "\n",
    "        # Entropy coef update part #####################################################\n",
    "        if entropy_coef_optimizer is not None:\n",
    "            # See Eq. (17) of the SAC and Applications paper\n",
    "            # log. probs have been computed when computing\n",
    "            # the actor loss\n",
    "            action_logprobs_rb = rb_workspace[\n",
    "                \"action_logprobs\"\n",
    "            ].detach()\n",
    "            entropy_coef_loss = -(\n",
    "                log_entropy_coef.exp() * (action_logprobs_rb + target_entropy)\n",
    "            ).mean()\n",
    "            entropy_coef_optimizer.zero_grad()\n",
    "            entropy_coef_loss.backward()\n",
    "            entropy_coef_optimizer.step()\n",
    "            logger.add_log(\"entropy_coef_loss\", entropy_coef_loss, agents.nb_steps)\n",
    "            logger.add_log(\"entropy_coef\", ent_coef, agents.nb_steps)\n",
    "\n",
    "        ####################################################\n",
    "\n",
    "        # Soft update of target q function\n",
    "        soft_update_params(agents.critic_1, agents.target_critic_1, tau)\n",
    "        soft_update_params(agents.critic_2, agents.target_critic_2, tau)\n",
    "        # soft_update_params(actor, target_actor, tau)\n",
    "        \n",
    "        agents.evaluate()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "6eb80540",
   "metadata": {},
   "source": [
    "## Definition of the parameters"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3de20637",
   "metadata": {},
   "outputs": [],
   "source": [
    "params={\n",
    "  \"save_best\": True,\n",
    "   \"base_dir\": \"${gym_env.env_name}/sac-S${algorithm.seed}_${current_time:}\",\n",
    "  \"logger\":{\n",
    "    \"classname\": \"bbrl.utils.logger.TFLogger\",\n",
    "    \"cache_size\": 10000,\n",
    "    \"every_n_seconds\": 10,\n",
    "    \"verbose\": False,    \n",
    "    },\n",
    "  \"algorithm\":{\n",
    "    \"seed\": 1,\n",
    "    \"n_envs\": 8,\n",
    "    \"n_steps\": 32,\n",
    "    \"buffer_size\": 1e6,\n",
    "    \"batch_size\": 256,\n",
    "    \"max_grad_norm\": 0.5,\n",
    "    \"nb_evals\": 16,\n",
    "    \"eval_interval\": 2_000,\n",
    "    \"learning_starts\": 10_000,\n",
    "    \"max_epochs\": 2_000,\n",
    "    \"discount_factor\": 0.98,\n",
    "    \"entropy_mode\": \"auto\", # \"auto\" or \"fixed\"\n",
    "    \"init_entropy_coef\": 2e-7,\n",
    "    \"tau_target\": 0.05,\n",
    "    \"architecture\":{\n",
    "      \"actor_hidden_size\": [64, 64],\n",
    "      \"critic_hidden_size\": [256, 256],\n",
    "    },\n",
    "    },\n",
    "    \"gym_env\":{\n",
    "        \"env_name\": \"CartPoleContinuous-v1\"\n",
    "        # \n",
    "    },\n",
    "  \"actor_optimizer\":{\n",
    "    \"classname\": \"torch.optim.Adam\",\n",
    "    \"lr\": 3e-4,\n",
    "    },\n",
    "  \"critic_optimizer\":{\n",
    "    \"classname\": \"torch.optim.Adam\",\n",
    "    \"lr\": 3e-4,\n",
    "    },\n",
    "  \"entropy_coef_optimizer\":{\n",
    "    \"classname\": \"torch.optim.Adam\",\n",
    "    \"lr\": 3e-4,\n",
    "    }\n",
    "}"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "b59b49e3",
   "metadata": {},
   "source": [
    "## Launching tensorboard to visualize the results"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "f5239750",
   "metadata": {},
   "outputs": [],
   "source": [
    "setup_tensorboard(\"./outputs/tblogs\")"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "184060bf",
   "metadata": {},
   "outputs": [],
   "source": [
    "agents = SACAgents(OmegaConf.create(params))\n",
    "run_sac(agents)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9596973f",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Visualize the best policy\n",
    "agents.visualize_best()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "5d431b34",
   "metadata": {},
   "source": [
    "## Exercises\n",
    "\n",
    "- use the same code on the Pendulum-v1 environment. This one is harder to\n",
    "  tune. Get the parameters from the\n",
    "  [rl-baseline3-zoo](https://github.com/DLR-RM/rl-baselines3-zoo) and see if\n",
    "  you manage to get SAC working on Pendulum"
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "cell_markers": "\"\"\""
  },
  "kernelspec": {
   "display_name": "Python 3 (ipykernel)",
   "language": "python",
   "name": "python3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
